@@ -6,7 +6,7 @@ use warnings;
 use Data::Dumper;
 use Carp;
 
-our $VERSION = '0.19_04';
+our $VERSION = '0.19_06';
 
 
 =head1 NAME
@@ -27,9 +27,7 @@ to install and activate the livestatus addon in your nagios installation.
 
 =head1 CONSTRUCTOR
 
-=over 4
-
-=item new ( [ARGS] )
+=head2 new ( [ARGS] )
 
 Creates an C<Nagios::MKLivestatus> object. C<new> takes at least the socketpath.
 Arguments are in key-value pairs.
@@ -43,12 +41,11 @@ Arguments are in key-value pairs.
     host_service_seperator    ascii code of the host/service seperator, defaults to 124 (pipe)
     keepalive                 enable keepalive. Default is off
     errors_are_fatal          errors will die with an error message. Default: on
+    timeout                   set a general timeout. Used for connect and querys, Default 10sec
 
 If the constructor is only passed a single argument, it is assumed to
 be a the C<socket> specification. Use either socker OR server.
 
-=back
-
 =cut
 
 sub new {
@@ -67,6 +64,7 @@ sub new {
                     "keepalive"                 => 0,       # enable keepalive?
                     "errors_are_fatal"          => 1,       # die on errors
                     "backend"                   => undef,   # should be keept undef, used internally
+                    "timeout"                   => 10,
                };
     bless $self, $class;
 
@@ -107,9 +105,7 @@ sub new {
 
 =head1 METHODS
 
-=over 4
-
-=item do
+=head2 do
 
  do($statement)
 
@@ -128,7 +124,7 @@ sub do {
 
 ########################################
 
-=item selectall_arrayref
+=head2 selectall_arrayref
 
  selectall_arrayref($statement)
  selectall_arrayref($statement, %opts)
@@ -199,7 +195,7 @@ sub selectall_arrayref {
 
 ########################################
 
-=item selectall_hashref
+=head2 selectall_hashref
 
  selectall_hashref($statement, $key_field)
 
@@ -232,7 +228,7 @@ sub selectall_hashref {
 
 ########################################
 
-=item selectcol_arrayref
+=head2 selectcol_arrayref
 
  selectcol_arrayref($statement)
  selectcol_arrayref($statement, %opt )
@@ -293,7 +289,7 @@ sub selectcol_arrayref {
 
 ########################################
 
-=item selectrow_array
+=head2 selectrow_array
 
  selectrow_array($statement)
 
@@ -316,7 +312,7 @@ sub selectrow_array {
 
 ########################################
 
-=item selectrow_arrayref
+=head2 selectrow_arrayref
 
  selectrow_arrayref($statement)
 
@@ -340,7 +336,7 @@ sub selectrow_arrayref {
 
 ########################################
 
-=item selectrow_hashref
+=head2 selectrow_hashref
 
  selectrow_hashref($statement)
 
@@ -364,7 +360,7 @@ sub selectrow_hashref {
 
 ########################################
 
-=item select_scalar_value
+=head2 select_scalar_value
 
  select_scalar_value($statement)
 
@@ -387,7 +383,7 @@ sub select_scalar_value {
 
 ########################################
 
-=item errors_are_fatal
+=head2 errors_are_fatal
 
  errors_are_fatal($values)
 
@@ -403,7 +399,29 @@ sub errors_are_fatal {
     $self->{'errors_are_fatal'}                = $value;
     $self->{'CONNECTOR'}->{'errors_are_fatal'} = $value;
 
-    return;
+    return 1;
+}
+
+
+########################################
+
+=head2 verbose
+
+ verbose($values)
+
+ Enable or disable verbose output. When enabled the module will dump out debug output
+
+ returns always true
+
+=cut
+sub verbose {
+    my $self  = shift;
+    my $value = shift;
+
+    $self->{'verbose'}                = $value;
+    $self->{'CONNECTOR'}->{'verbose'} = $value;
+
+    return 1;
 }
 
 
@@ -416,72 +434,98 @@ sub _send {
     my $statement = shift;
     my $header = "";
 
-    undef $Nagios::MKLivestatus::ErrorCode;
+    $Nagios::MKLivestatus::ErrorCode = 0;
     undef $Nagios::MKLivestatus::ErrorMessage;
 
-    croak("no statement") if !defined $statement;
+    return(490, $self->_get_error(490), undef) if !defined $statement;
     chomp($statement);
 
+    my($status,$msg,$body);
     if($statement =~ m/^Separators:/mx) {
-        $Nagios::MKLivestatus::ErrorCode    = 492;
-        $Nagios::MKLivestatus::ErrorMessage = "Separators not allowed in statement. Please use the seperator options in new()";
-        return;
+        $status = 492;
+        $msg    = $self->_get_error($status);
     }
 
-    if($statement =~ m/^KeepAlive:/mx) {
-        $Nagios::MKLivestatus::ErrorCode    = 496;
-        $Nagios::MKLivestatus::ErrorMessage = "Keepalive not allowed in statement. Please use the keepalive option in new()";
-        return;
+    elsif($statement =~ m/^KeepAlive:/mx) {
+        $status = 496;
+        $msg    = $self->_get_error($status);
     }
 
-    if($statement =~ m/^ResponseHeader:/mx) {
-        $Nagios::MKLivestatus::ErrorCode    = 495;
-        $Nagios::MKLivestatus::ErrorMessage = "ResponseHeader not allowed in statement. Header will be set automatically";
-        return;
+    elsif($statement =~ m/^ResponseHeader:/mx) {
+        $status = 495;
+        $msg    = $self->_get_error($status);
     }
 
-    if($statement =~ m/^ColumnHeaders:/mx) {
-        $Nagios::MKLivestatus::ErrorCode    = 494;
-        $Nagios::MKLivestatus::ErrorMessage = "ColumnHeaders not allowed in statement. Header will be set automatically";
-        return;
+    elsif($statement =~ m/^ColumnHeaders:/mx) {
+        $status = 494;
+        $msg    = $self->_get_error($status);
     }
 
-    if($statement =~ m/^OuputFormat:/mx) {
-        $Nagios::MKLivestatus::ErrorCode    = 493;
-        $Nagios::MKLivestatus::ErrorMessage = "OuputFormat not allowed in statement. Header will be set automatically";
-        return;
+    elsif($statement =~ m/^OuputFormat:/mx) {
+        $status = 493;
+        $msg    = $self->_get_error($status);
     }
 
-    # Commands need no additional header
-    if($statement !~ m/^COMMAND/mx) {
-        $header .= "Separators: $self->{'line_seperator'} $self->{'column_seperator'} $self->{'list_seperator'} $self->{'host_service_seperator'}\n";
-        $header .= "ResponseHeader: fixed16\n";
+    # should be cought in mlivestatus directly
+    elsif($statement =~ m/^Limit:\ (.*)$/mx and $1 !~ m/^\d+$/mx) {
+        $status = 403;
+        $msg    = $self->_get_error($status);
     }
-    if($self->{'keepalive'}) {
-        $header .= "KeepAlive: on\n";
-    }
-    my $send = "$statement\n$header";
-    print "> ".Dumper($send) if $self->{'verbose'};
-    my($status,$msg,$body) = $self->_send_socket($send);
-    if($self->{'verbose'}) {
-        print "status: ".Dumper($status);
-        print "msg:    ".Dumper($msg);
-        print "< ".Dumper($body);
+    elsif($statement =~ m/^GET\ (.*)$/mx and $1 =~ m/^\s*$/mx) {
+        $status = 403;
+        $msg    = $self->_get_error($status);
     }
 
-    if($status != 200) {
+    elsif($statement =~ m/^Columns:\ (.*)$/mx and ($1 =~ m/,/mx or $1 =~ /^\s*$/mx)) {
+        $status = 405;
+        $msg    = $self->_get_error($status);
+    }
+    elsif($statement !~ m/^GET\ /mx and $statement !~ m/^COMMAND\ /mx) {
+        $status = 401;
+        $msg    = $self->_get_error($status);
+    }
+
+    else {
+
+        # Commands need no additional header
+        if($statement !~ m/^COMMAND/mx) {
+            $header .= "Separators: $self->{'line_seperator'} $self->{'column_seperator'} $self->{'list_seperator'} $self->{'host_service_seperator'}\n";
+            $header .= "ResponseHeader: fixed16\n";
+            #$header .= "ColumnHeaders: on\n";
+
+        }
+        if($self->{'keepalive'}) {
+            $header .= "KeepAlive: on\n";
+        }
+        my $send = "$statement\n$header";
+        print "> ".Dumper($send) if $self->{'verbose'};
+        ($status,$msg,$body) = $self->_send_socket($send);
+        if($self->{'verbose'}) {
+            print "status: ".Dumper($status);
+            print "msg:    ".Dumper($msg);
+            print "< ".Dumper($body);
+        }
+    }
+
+    if($status >= 300) {
         $body = '' if !defined $body;
         chomp($body);
-        $msg = "$msg\n$body";
         $Nagios::MKLivestatus::ErrorCode    = $status;
-        $Nagios::MKLivestatus::ErrorMessage = $msg;
+        if(defined $body and $body ne '') {
+            $Nagios::MKLivestatus::ErrorMessage = $body;
+        } else {
+            $Nagios::MKLivestatus::ErrorMessage = $msg;
+        }
         if($self->{'errors_are_fatal'}) {
-            croak("ERROR ".$status." - ".$msg."\nin query:\n".$statement);
+            my $nicestatement = $statement;
+            $nicestatement    =~ s/\n/\\n/gmx;
+            croak("ERROR ".$status." - ".$Nagios::MKLivestatus::ErrorMessage." in query:\n'".$nicestatement."'\n");
         }
         return;
     }
 
-    return if !defined $body;
+    # return a empty result set if nothing found
+    return({ keys => [], result => []}) if !defined $body;
 
     my $line_seperator = chr($self->{'line_seperator'});
     my $col_seperator  = chr($self->{'column_seperator'});
@@ -540,9 +584,7 @@ sub _send_socket {
     my $statement = shift;
     my($recv,$header);
 
-    croak("no statement") if !defined $statement;
-
-    my $sock = $self->_open() or return(491, 'failed to connect', undef);
+    my $sock = $self->_open() or return(491, $self->_get_error(491), $!);
     print $sock $statement;
     if($self->{'keepalive'}) {
         print $sock "\n";
@@ -553,19 +595,11 @@ sub _send_socket {
     # COMMAND statements never return something
     if($statement =~ m/^COMMAND/mx) {
         $self->_close($sock) unless $self->{'keepalive'};
-        return('200', 'COMMANDs never return something', undef);
-    }
-
-    if(!$sock->opened()) {
-        confess("socket is not open, cannot read");
-    }
-
-    if($sock->error()) {
-        confess("socket has errors, cannot read");
+        return('201', $self->_get_error(201), undef);
     }
 
     $sock->read($header, 16) or return($self->_socket_error($statement, $sock, 'reading header from socket failed'));
-    print "header: $header\n" if $self->{'verbose'};
+    print "header: $header" if $self->{'verbose'};
     my($status, $msg, $content_length) = $self->_parse_header($header);
     return($status, $msg, undef) if !defined $content_length;
     if($content_length > 0) {
@@ -592,7 +626,7 @@ sub _socket_error {
     } else {
         carp($message);
     }
-    return(500, 'socket error', undef);
+    return(500, $self->_get_error(500), undef);
 }
 
 ########################################
@@ -601,42 +635,73 @@ sub _parse_header {
     my $header = shift;
 
     if(!defined $header) {
-        return(497, 'got no header', undef);
+        return(497, $self->_get_error(497), undef);
     }
 
     my $headerlength = length($header);
     if($headerlength != 16) {
-        return(498, 'header is not exactly 16byte long', undef);
+        return(498, $self->_get_error(498), undef);
     }
     chomp($header);
 
     my $status         = substr($header,0,3);
     my $content_length = substr($header,5);
     if($content_length !~ m/^\s*(\d+)$/mx) {
-        return(499, 'failed to get content-length from header', undef);
+        return(499, $self->_get_error(499), undef);
     } else {
         $content_length = $1;
     }
 
-    #print "status: ".$status."\n";
-    #print "length: ".$content_length."\n";
+    return($status, $self->_get_error($status), $content_length);
+}
+
+########################################
+
+=head1 ERROR HANDLING
+
+Errorhandling can be done like this:
+
+    use Nagios::MKLivestatus;
+    my $nl = Nagios::MKLivestatus->new( socket => '/var/lib/nagios3/rw/livestatus.sock' );
+    $nl->errors_are_fatal(0);
+    my $hosts = $nl->selectall_arrayref("GET hosts");
+    if($Nagios::MKLivestatus::ErrorCode){
+        croak($Nagios::MKLivestatus::ErrorMessage);
+    }
+
+=cut
+sub _get_error {
+    my $self = shift;
+    my $code = shift;
 
     my $codes = {
         '200' => 'OK. Reponse contains the queried data.',
+        '201' => 'COMMANDs never return something',
         '401' => 'The request contains an invalid header.',
         '402' => 'The request is completely invalid.',
         '403' => 'The request is incomplete.',
         '404' => 'The target of the GET has not been found (e.g. the table).',
         '405' => 'A non-existing column was being referred to',
+        '490' => 'no query',
+        '491' => 'failed to connect',
+        '492' => 'Separators not allowed in statement. Please use the seperator options in new()',
+        '493' => 'OuputFormat not allowed in statement. Header will be set automatically',
+        '494' => 'ColumnHeaders not allowed in statement. Header will be set automatically',
+        '495' => 'ResponseHeader not allowed in statement. Header will be set automatically',
+        '496' => 'Keepalive not allowed in statement. Please use the keepalive option in new()',
+        '497' => 'got no header',
+        '498' => 'header is not exactly 16byte long',
+        '499' => 'failed to get content-length from header',
+        '500' => 'socket error',
     };
 
-    return($status, $codes->{$status}, $content_length);
+    confess('non existant error code: '.$code) if !defined $codes->{$code};
+
+    return($codes->{$code});
 }
 
 1;
 
-=back
-
 =head1 SEE ALSO
 
 For more information about the query syntax and the livestatus plugin installation
